/****************************************************************************/
// camcam.c
//
//  - camera init & shutdown
//  - camera disconnected/reconnected notifications
//
/****************************************************************************/

/* ***** BEGIN LICENSE BLOCK *****
 * Version: MPL 1.1
 *
 * The contents of this file are subject to the Mozilla Public License Version
 * 1.1 (the "License"); you may not use this file except in compliance with
 * the License. You may obtain a copy of the License at
 * http://www.mozilla.org/MPL/
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 * for the specific language governing rights and limitations under the
 * License.
 *
 * The Original Code is Cameraderie.
 *
 * The Initial Developer of the Original Code is Richard L. Walsh.
 * 
 * Portions created by the Initial Developer are Copyright (C) 2006,2007
 * the Initial Developer. All Rights Reserved.
 *
 * Contributor(s):
 *
 * ***** END LICENSE BLOCK ***** */

/****************************************************************************/

#define INCL_DOS
#define INCL_WIN
#include "camapp.h"
#include "camusb.h"
#include <os2.h>

/****************************************************************************/

enum    eNotify     { eSEMADD, eSEMREMOVE, eSEMTERM, eCNTSEMS, eSEMERR};

typedef struct _CAMNotify {
    SEMRECORD   sr[eCNTSEMS];
    CAMPtr      thisCam;
    HMUX        hmux;
    HWND        hwnd;
    uint32_t    NotifyID;
} CAMNotify;

/****************************************************************************/

ULONG           CamSetupCamera( HMQ hmq, void ** thisCamPtr);
void            CamShutDownCamera( void ** thisCamPtr);
ULONG           CamReconnectCamera( CAMPtr thisCam);
ULONG           CamAbandonCamera( CAMPtr thisCam);
void            CamGetSupportedOps( CAMPtr thisCam);
ULONG           CamInitNotification( HWND hwnd, void * voidCam);
void            CamNotifyThreadProc( void * arg);
void APIENTRY   CamTermNotifyOnExit( ULONG ul);
void            CamHandleNotification( HWND hwnd, MPARAM mp1, MPARAM mp2);
ULONG           CamShowInitError( ULONG ulErr, ULONG ulDlgID);
MRESULT _System CamInitErrorDlgProc( HWND hwnd, ULONG msg, MPARAM mp1, MPARAM mp2);
MRESULT _System CamInitErrorDlgProc( HWND hwnd, ULONG msg, MPARAM mp1, MPARAM mp2);

/****************************************************************************/

static ULONG            ulNotifyID = 0;

static CAMNLS   nlsNoCamDlg[] = {
    { IDC_TEXT1,        DLG_NoCameras },
    { IDC_TEXT2,        DLG_PlugIn },
    { IDC_TEXT4,    	DLG_ToEndNow },
    { IDC_RETRY,    	BTN_Retry },
    { IDC_EXIT,     	BTN_Exit },
    { 0,                0 }};

static CAMNLS   nlsErrorDlg[] = {
    { IDC_TEXT4,    	DLG_ToEnd },
    { IDC_EXIT,     	BTN_Exit },
    { 0,                0 }};

/****************************************************************************/

ULONG       CamSetupCamera( HMQ hmq, void ** thisCamPtr)

{
    ULONG   	     rc;
    ULONG            ctr;
    CAMPtr           thisCam = 0;
    CAMCameraListPtr pList = 0;
    CAMDevice        device;

do
{
    *thisCamPtr = 0;
    SetUsbDebug( 2);
    rc = InitUsb();
    if (rc)
        break;

    thisCam = (CAMPtr)malloc( sizeof(CAM));
    if (!thisCam) {
        rc = CAMERR_MALLOC;
        break;
    }
    memset( thisCam, 0, sizeof(CAM));
    *thisCamPtr = thisCam;

    // create the camera-access mutex in an owned state
    rc = DosCreateMutexSem( 0, (PHMTX)(&thisCam->hmtxCam), 0, TRUE);
    if (rc)
        break;

    rc = GetCameraList( &pList, (uint32_t)CAM_FORCE_FLAG);

    while (rc == CAMERR_NOCAMERAS) {
        if (CamShowInitError( 0, IDD_NOCAM) != IDC_RETRY)
            break;

        rc = GetCameraList( &pList, (uint32_t)CAM_FORCE_FLAG);
    }
    if (rc)
        break;

    // if the force flag is off, there's only 1 entry in the list;
    // if it's on, try all found devices
    rc = CAMERR_USBFAILURE;
    for (ctr = 0; ctr < pList->cntCameras; ctr++) {
        device = pList->Cameras[ctr].Device;
        rc = InitCamera( device, &thisCam->hCam);
        if (!rc)
            break;
    }
    if (rc)
        break;

    thisCam->pUsbInfo = (CAMUsbInfo*)malloc( sizeof(CAMUsbInfo));
    if (!thisCam->pUsbInfo) {
        rc = CAMERR_MALLOC;
        break;
    }
    memcpy( thisCam->pUsbInfo, &pList->Cameras[ctr], sizeof(CAMUsbInfo));

    rc = GetDeviceInfo( thisCam->hCam, &thisCam->pDevInfo);
    if (rc)
        break;

    CamGetSupportedOps( thisCam);

} while (0);

    if (thisCam && thisCam->hmtxCam)
        DosReleaseMutexSem( thisCam->hmtxCam);

    // if there was an error, abort
    if (rc && rc != CAMERR_NOCAMERAS)
        CamShowInitError( MSG_InitCamera, IDD_ERROR);

    free( pList);

    return rc;
}

/****************************************************************************/

void        CamShutDownCamera( void ** thisCamPtr)

{
    if (*thisCamPtr) {
        CAMPtr  thisCam = (CAMPtr)*thisCamPtr;

        if (thisCam->hevTerm) {
            DosPostEventSem( thisCam->hevTerm);
            DosSleep( 30);
        }

        free( thisCam->pUsbInfo);
        free( thisCam->pDevInfo);
        thisCam->pUsbInfo = 0;
        thisCam->pDevInfo = 0;

        // don't worry if this times-out
        if (thisCam->hmtxCam)
            DosRequestMutexSem( thisCam->hmtxCam, CAM_MUTEXWAIT);

        TermCamera( &thisCam->hCam);

        if (thisCam->hmtxCam) {
            DosReleaseMutexSem( thisCam->hmtxCam);
            DosCloseMutexSem( thisCam->hmtxCam);
        }

        free( thisCam);
        *thisCamPtr = 0;
    }
    TermUsb();

    return;
}

/****************************************************************************/

ULONG       CamReconnectCamera( CAMPtr thisCam)

{
    ULONG   	 rc;

do {
    if (!thisCam || !thisCam->pUsbInfo)
        return (CAMERR_NOCAMERAS);

    rc = DosRequestMutexSem( thisCam->hmtxCam, CAM_MUTEXWAIT);
    if (rc)
        break;

    // when the camera is disconnected, the notify thread zeros hCam
    // but saves its address in hDisconnect so it can be freed when
    // nothing is using it
    if (thisCam->hDisconnect) {
        free( thisCam->hDisconnect);
        thisCam->hDisconnect = 0;
    }

    rc = GetSpecificCamera( thisCam->pUsbInfo);
    if (rc)
        break;

    rc = InitCamera( thisCam->pUsbInfo->Device, &thisCam->hCam);
    if (rc)
        break;

    rc = GetDeviceInfo( thisCam->hCam, &thisCam->pDevInfo);
    if (rc)
        break;

    CamGetSupportedOps( thisCam);

} while (0);

    DosReleaseMutexSem( thisCam->hmtxCam);

    return rc;
}

/***************************************************************************/

ULONG       CamAbandonCamera( CAMPtr thisCam)

{
    if (!thisCam || !thisCam->pUsbInfo)
        return (CAMERR_NOCAMERAS);

    // note that hCam (actually, hDisconnect) isn't freed until the
    // camera is reconnected;  this avoids blocking the msg queue
    // while waiting for the mutex

    free( thisCam->pDevInfo);
    thisCam->pDevInfo = 0;

    thisCam->pUsbInfo->Device = 0;
    thisCam->pUsbInfo->BusName = 0;
    thisCam->pUsbInfo->DeviceNbr = 0;

    return 0;
}

/****************************************************************************/

void        CamGetSupportedOps( CAMPtr thisCam)

{
    uint32_t        ctr;
    uint32_t        cnt;
    uint32_t *      ops;

do {
    if (!thisCam->pDevInfo)
        break;

    ops = thisCam->pDevInfo->OperationsSupported;
    if (!ops)
        break;

    thisCam->camFlags &= ~CAM_PARTIAL_MASK;
    cnt = thisCam->pDevInfo->cntOperationsSupported;

    for (ctr = 0; ctr < cnt; ops++, ctr++)
        if (*ops == PTP_OC_GetPartialObject)
            thisCam->camFlags |= CAM_PARTIAL_STANDARD;
        else
        if (thisCam->pUsbInfo->VendorId == 0x04A9 &&
            *ops == PTP_OC_CANON_GetPartialObject)
            thisCam->camFlags |= CAM_PARTIAL_CANON;

} while (0);

    return;
}

/***************************************************************************/
/****************************************************************************/

ULONG       CamInitNotification( HWND hwnd, void * voidCam)

{
    APIRET      rc = 0;
    CAMNotify * pCN = 0;
    CAMPtr      thisCam = (CAMPtr)voidCam;

do {
    pCN = malloc( sizeof(CAMNotify));
    if (!pCN) {
        rc = CAMERR_MALLOC;
        break;
    }
    memset( pCN, 0, sizeof(CAMNotify));

    pCN->hwnd = hwnd;
    pCN->thisCam = thisCam;
    if (!thisCam->pUsbInfo) {
        rc = CAMERR_MISCELLANEOUS;
        break;
    }

    pCN->sr[eSEMADD].ulUser = eSEMADD;
    rc = DosCreateEventSem( 0, (PHEV)&(pCN->sr[eSEMADD].hsemCur), DC_SEM_SHARED, 0);
    if (rc)
        break;

    pCN->sr[eSEMREMOVE].ulUser = eSEMREMOVE;
    rc = DosCreateEventSem( 0, (PHEV)&(pCN->sr[eSEMREMOVE].hsemCur), DC_SEM_SHARED, 0);
    if (rc)
        break;

    pCN->sr[eSEMTERM].ulUser = eSEMTERM;
    rc = DosCreateEventSem( 0, (PHEV)&(pCN->sr[eSEMTERM].hsemCur), 0, 0);
    if (rc)
        break;

    thisCam->hevTerm = (HEV)pCN->sr[eSEMTERM].hsemCur;

    rc = DosCreateMuxWaitSem( 0, &pCN->hmux, eCNTSEMS, pCN->sr, DCMW_WAIT_ANY);
    if (rc)
        break;

    rc = UsbStartNotification( thisCam->pUsbInfo, (HEV)pCN->sr[eSEMADD].hsemCur,
                            (HEV)pCN->sr[eSEMREMOVE].hsemCur, &pCN->NotifyID);
    if (rc)
        break;

    ulNotifyID = pCN->NotifyID;
    rc = DosExitList( ((0x7f << 8) | EXLST_ADD), CamTermNotifyOnExit);
    if (rc) {
        camcli_error( "InitNotification:  DosExitList ADD failed - rc= %d", (int)rc);
        rc = 0;
    }

    if (_beginthread( CamNotifyThreadProc, 0, 0x4000, pCN) == -1) {
        rc = CAMERR_MISCELLANEOUS;
        break;
    }

} while (0);

    if (rc) {
        camcli_error( "CamInitNotification - rc= 0x%x\n", rc);

        if (pCN) {
            if (pCN->NotifyID)
                EndNotification( pCN->NotifyID);
            if (pCN->hmux)
                DosCloseMuxWaitSem( pCN->hmux);
            if (pCN->sr[eSEMTERM].hsemCur)
                DosCloseEventSem( (HEV)pCN->sr[eSEMTERM].hsemCur);
            if (pCN->sr[eSEMREMOVE].hsemCur)
                DosCloseEventSem( (HEV)pCN->sr[eSEMREMOVE].hsemCur);
            if (pCN->sr[eSEMADD].hsemCur)
                DosCloseEventSem( (HEV)pCN->sr[eSEMADD].hsemCur);
            free( pCN);
        }
        thisCam->hevTerm = 0;
    }

    return (rc);
}

/****************************************************************************/

void        CamNotifyThreadProc( void * arg)

{
    APIRET      rc;
    CAMNotify * pCN = (CAMNotify*)arg;
    ULONG       ulSem;
    ULONG       ulCnt;

    if (!arg)
        return;

    DosResetEventSem( (HEV)pCN->sr[eSEMADD].hsemCur, &ulCnt);

    while (1) {
        rc = DosWaitMuxWaitSem( pCN->hmux, (ULONG)-1, &ulSem);
        if (rc)
            ulSem = eSEMERR;

        if (ulSem < eCNTSEMS) {
            ulCnt = 0;
            rc = DosResetEventSem( (HEV)pCN->sr[ulSem].hsemCur, &ulCnt);
            if (ulCnt != 1)
                camcli_error( "Multiple event sem postings - sem= %d  cnt= %d",
                              (int)ulSem, (int)ulCnt);
            if (rc)
                ulSem = eSEMERR;
        }
        else
            ulSem = eSEMERR;

        // if this is a disconnect, make hCam invalid immediately
        // but save its address so it can be freed by the primary
        // thread when it's safe to do so
        if (ulSem != eSEMTERM) {
            pCN->thisCam->hDisconnect = pCN->thisCam->hCam;
            pCN->thisCam->hCam = 0;
            WinPostMsg( pCN->hwnd, CAMMSG_NOTIFY, (MP)ulSem, (MP)pCN->thisCam);
        }
        if (rc || (ulSem != eSEMADD && ulSem != eSEMREMOVE))
            break;
    }

    if (rc)
        camcli_error( "exiting CamNotifyThreadProc - rc= %d", (int)rc);

    rc = EndNotification( pCN->NotifyID);
    if (rc)
        camcli_error( "EndNotification - rc= %x", (int)rc);
    else {
        ulNotifyID = 0;
        rc = DosExitList( EXLST_REMOVE, CamTermNotifyOnExit);
        if (rc)
            camcli_error( "DosExitList REMOVE failed - rc= %d", (int)rc);
    }

    DosCloseMuxWaitSem( pCN->hmux);
    pCN->thisCam->hevTerm = 0;
    DosCloseEventSem( (HEV)pCN->sr[eSEMTERM].hsemCur);
    DosCloseEventSem( (HEV)pCN->sr[eSEMREMOVE].hsemCur);
    DosCloseEventSem( (HEV)pCN->sr[eSEMADD].hsemCur);
    free( pCN);

    return;
}

/****************************************************************************/

// this should clean up in case we crash

void APIENTRY   CamTermNotifyOnExit( ULONG ul)

{
    if (ulNotifyID) {
        APIRET rc = EndNotification( ulNotifyID);
        printf( "CamTermNotifyOnExit:  TermNotification - rc= %x\n", (int)rc);
        ulNotifyID = 0;
    }

    DosExitList( EXLST_EXIT, 0);
    return;
}

/****************************************************************************/

void        CamHandleNotification( HWND hwnd, MPARAM mp1, MPARAM mp2)

{
    ULONG   rc;
    HWND    hDlg;

    if (mp1 == (MP)eSEMREMOVE) {
        OpenWindow( hwnd, IDD_DISCONNECT);
        CamAbandonCamera( (CAMPtr)mp2);

        return;
    }

    if (mp1 == (MP)eSEMADD) {
        rc = CamReconnectCamera( (CAMPtr)mp2);

        hDlg = GetHwnd( IDD_DISCONNECT);
        if (hDlg) {
            WindowClosed( IDD_DISCONNECT);
            WinDestroyWindow( hDlg);
        }

        if (!rc)
            WinPostMsg( hwnd, WM_COMMAND, (MP)IDM_SYNCDATA, 0);
        else {
            CamShowInitError( MSG_ReopenCamera, IDD_ERROR);
            WinPostMsg( hwnd, WM_QUIT, 0, 0);
        }

        return;
    }

    if (mp2) {
        if (CamInitNotification( hwnd, (CAMPtr)mp2))
            WinAlarm( HWND_DESKTOP, WA_ERROR);
    }

    return;
}

/****************************************************************************/
/****************************************************************************/

// this is only used before msg loop processing begins

ULONG       CamShowInitError( ULONG ulErr, ULONG ulDlgID)

{
    char    szText[256];

    if (ulErr)
        LoadString( ulErr, szText);

    return (WinDlgBox( HWND_DESKTOP,        // parent-window
                       NULLHANDLE,          // owner-window
                       CamInitErrorDlgProc, // dialog proc
                       NULLHANDLE,          // EXE module handle
                       ulDlgID,             // dialog id
                       (ulErr ? szText : 0)));  // pointer to create params
}

/****************************************************************************/

MRESULT _System CamInitErrorDlgProc( HWND hwnd, ULONG msg, MPARAM mp1, MPARAM mp2)
 
{
    RECTL   rcl;

    if (msg != WM_INITDLG)
        return (WinDefDlgProc( hwnd, msg, mp1, mp2));

    if (mp2)
        WinSetDlgItemText( hwnd, IDC_TEXT1, (char*)mp2);

    if (WinQueryWindowUShort( hwnd, QWS_ID) == IDD_NOCAM)
        LoadDlgStrings( hwnd, nlsNoCamDlg);
    else
        LoadDlgStrings( hwnd, nlsErrorDlg);

    WinQueryWindowRect( hwnd, &rcl);
    rcl.xLeft = WinQuerySysValue( HWND_DESKTOP, SV_CXSCREEN);
    rcl.yBottom = WinQuerySysValue( HWND_DESKTOP, SV_CYSCREEN);

    WinSetWindowPos( hwnd, 0,
                     (rcl.xLeft - rcl.xRight) / 2,
                     (rcl.yBottom - rcl.yTop) / 2,
                     0, 0, SWP_MOVE | SWP_SHOW);

    return 0;
}

/****************************************************************************/

