/****************************************************************************/
// camjpg.c
//
//  - routines to rotate a JPEG while saving it
//  - routines to determine a JPEG's orientation when loading it
//
//  Much of the source code in this file was taken from libjpeg.
//  - SaveRotatedJPEG() was taken almost verbatim from jpegtran.c;
//  - the camjpg_* routines are derived from similar code in
//    jdatasrc.c and jdatadst.c.
//  As required by the license for libjpeg:
//
//  This software is based in part on the work of the Independent JPEG Group.
//      Copyright (C) 1994-1997, Thomas G. Lane.
//  For conditions of distribution and use, see the accompanying
//  readme-jpg.txt file.
//
/****************************************************************************/

/* ***** BEGIN LICENSE BLOCK *****
 * Version: MPL 1.1
 *
 * The contents of this file are subject to the Mozilla Public License Version
 * 1.1 (the "License"); you may not use this file except in compliance with
 * the License. You may obtain a copy of the License at
 * http://www.mozilla.org/MPL/
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 * for the specific language governing rights and limitations under the
 * License.
 *
 * The Original Code is Cameraderie.
 *
 * The Initial Developer of the Original Code is Richard L. Walsh.
 * 
 * Portions created by the Initial Developer are Copyright (C) 2006,2007
 * the Initial Developer. All Rights Reserved.
 *
 * Contributor(s):
 *  Thomas G. Lane, Independent JPEG Group - Copyright (C) 1994-1997
 *
 * ***** END LICENSE BLOCK ***** */

/****************************************************************************/
/****************************************************************************/

// from jpegtran.c
#include "cdjpeg.h"         // Common decls for cjpeg/djpeg applications
#include "transupp.h"       // Support routines for jpegtran
#include "jversion.h"       // for version message

// from jdatadst.c & jdatasrc.c
#include "jinclude.h"
#include "jpeglib.h"
#include "jerror.h"

// Cam includes
#include <setjmp.h>
#include "camusb.h"
#define INCL_DOS
#include <os2.h>

/****************************************************************************/
/****************************************************************************/

typedef struct _CAMJpg {
    CAMSaveRot *    csr;
    JOCTET *        pInbuf;                 // start of buffer
    uint32_t        cbInbuf;                // size of buffer
    uint32_t    	cbToGet;                // bytes to fetch
    JOCTET *        pOutbuf;                // start of buffer
    uint32_t        cbOutbuf;               // size of buffer
    jmp_buf         jmpbuf;                 // setjmp buffer
} CAMJpg;

#define OUTPUT_BUF_SIZE     0x100000

/****************************************************************************/

typedef struct _CAMJThumbInt {
    CAMJThumb *     cjt;
    uint32_t        rtn;
    uint32_t        eoiCtr;
    JOCTET          eoi[4];
    jmp_buf         jmpbuf;
} CAMJThumbInt;

/****************************************************************************/

#define M_SOI   0xD8
#define M_EOI   0xD9
#define M_SOS   0xDA
#define M_APP1	0xE1

/****************************************************************************/

uint32_t        JpgSaveRotatedJPEG( CAMSaveRot * csr);
void            JSavRot_stdio_src( j_decompress_ptr cinfo);
void            JSavRot_init_source( j_decompress_ptr cinfo);
boolean         JSavRot_fill_input_buffer( j_decompress_ptr cinfo);
void            JSavRot_skip_input_data( j_decompress_ptr cinfo, long num_bytes);
void            JSavRot_term_source( j_decompress_ptr cinfo);
uint32_t        JSavRot_read_input( j_decompress_ptr cinfo);
void            JSavRot_stdio_dest( j_compress_ptr cinfo);
void            JSavRot_init_destination( j_compress_ptr cinfo);
boolean         JSavRot_empty_output_buffer( j_compress_ptr cinfo);
void            JSavRot_term_destination( j_compress_ptr cinfo);
void            JSavRot_error_exit (j_common_ptr cinfo);
void            JSavRot_abort( j_common_ptr cinfo);

uint32_t        JpgDecodeThumbnail( CAMJThumb * cjt);
void            JThumb_stdio_src( j_decompress_ptr cinfo);
void            JThumb_init_source( j_decompress_ptr cinfo);
boolean         JThumb_fill_input_buffer( j_decompress_ptr cinfo);
void            JThumb_skip_input_data( j_decompress_ptr cinfo, long num_bytes);
void            JThumb_term_source( j_decompress_ptr cinfo);
void            JThumb_error_exit( j_common_ptr cinfo);
void            JThumb_abort( j_common_ptr cinfo);

uint32_t        JpgFindOrientation( char * buf, uint32_t count, uint32_t* orientation);
uint32_t        JpgFindExifTiffHdr( char * buf, char * end, char** ppHdr);
uint32_t        JpgFind0thIfd( char * hdr, char * end, char** ppIFD, uint32_t* fLE);

/****************************************************************************/
/****************************************************************************/

uint32_t    JpgSaveRotatedJPEG( CAMSaveRot * csr)

{
    struct jpeg_decompress_struct   srcinfo;
    struct jpeg_compress_struct     dstinfo;
    struct jpeg_error_mgr           jsrcerr;
    struct jpeg_error_mgr           jdsterr;
    jpeg_transform_info             transformoption;
    jvirt_barray_ptr *              src_coef_arrays;
    jvirt_barray_ptr *              dst_coef_arrays;
    CAMJpg                          cjpg;

do {
    memset( &cjpg, 0, sizeof(CAMJpg));
    cjpg.csr = csr;
    if (setjmp( cjpg.jmpbuf))
        break;

    // Initialize the JPEG decompression object with custom error handling.
    srcinfo.client_data = &cjpg;
    srcinfo.err = jpeg_std_error( &jsrcerr);
    jsrcerr.error_exit = &JSavRot_error_exit;
    jpeg_create_decompress( &srcinfo);

    // Initialize the JPEG compression object with custom error handling.
    dstinfo.client_data = &cjpg;
    dstinfo.err = jpeg_std_error( &jdsterr);
    jdsterr.error_exit = &JSavRot_error_exit;
    jpeg_create_compress( &dstinfo);

    jsrcerr.trace_level = jdsterr.trace_level;
    srcinfo.mem->max_memory_to_use = dstinfo.mem->max_memory_to_use;

    // Specify data source for decompression
    JSavRot_stdio_src( &srcinfo);

    // Enable saving of extra markers that we want to copy
    jcopy_markers_setup( &srcinfo, JCOPYOPT_ALL);

    // Read file header
    jpeg_read_header( &srcinfo, TRUE);

    // set transform options
    transformoption.trim = FALSE;
    transformoption.force_grayscale = FALSE;
    if (csr->ulRotate == 90)
        transformoption.transform = JXFORM_ROT_90;
    else
    if (csr->ulRotate == 180)
        transformoption.transform = JXFORM_ROT_180;
    else
        transformoption.transform = JXFORM_ROT_270;

    // Any space needed by a transform option must be requested before
    // jpeg_read_coefficients so that memory allocation will be done right.
    jtransform_request_workspace( &srcinfo, &transformoption);

    // Read source file as DCT coefficients
    src_coef_arrays = jpeg_read_coefficients( &srcinfo);

    // Initialize destination compression parameters from source values
    jpeg_copy_critical_parameters( &srcinfo, &dstinfo);

    // Adjust destination parameters if required by transform options;
    // also find out which set of coefficient arrays will hold the output.
    dst_coef_arrays = jtransform_adjust_parameters( &srcinfo, &dstinfo,
                                    src_coef_arrays, &transformoption);

    // Specify data destination for compression
    JSavRot_stdio_dest( &dstinfo);
  
    // Start compressor (note no image data is actually written here)
    jpeg_write_coefficients( &dstinfo, dst_coef_arrays);
  
    // Copy to the output file any extra markers that we want to preserve
    jcopy_markers_execute( &srcinfo, &dstinfo, JCOPYOPT_ALL);
  
    // Execute image transformation
    jtransform_execute_transformation( &srcinfo, &dstinfo,
                                       src_coef_arrays, &transformoption);
  
    // Finish compression and release memory
    jpeg_finish_compress( &dstinfo);
    (void) jpeg_finish_decompress( &srcinfo);

} while (0);

    jpeg_destroy_compress( &dstinfo);
    jpeg_destroy_decompress( &srcinfo);

    if (cjpg.pInbuf)
        DosFreeMem( cjpg.pInbuf);
    if (cjpg.pOutbuf)
        DosFreeMem( cjpg.pOutbuf);

    // All done
    return (csr->rtn);
}

/****************************************************************************/
/****************************************************************************/

void        JSavRot_stdio_src( j_decompress_ptr cinfo)

{
    uint32_t        rc;
    uint32_t        cnt;
    CAMJpg *        cjpg = (CAMJpg *)cinfo->client_data;
    CAMSaveRot *    csr = cjpg->csr;
    CAMReadData     crd;

do {
    if (*csr->pStop) {
        csr->rtn = CAMERR_REQCANCELLED;
        break;
    }

    csr->rtn = ReadObjectBegin( csr->thisCam->hCam, csr->hObject, &cjpg->cbToGet, &crd);
    if (csr->rtn != PTP_RC_OK)
        break;

    if (cjpg->cbToGet <= sizeof(crd.data)) {
        csr->rtn = ReadObjectEnd( csr->thisCam->hCam);
        if (csr->rtn != PTP_RC_OK)
            break;
        cjpg->cbInbuf = cjpg->cbToGet;
        cnt = cjpg->cbToGet;
        cjpg->cbToGet = 0;
    }
    else {
        cjpg->cbInbuf = (cjpg->cbToGet < 0x100000) ? cjpg->cbToGet : 0x100000;
        cnt = sizeof(crd.data);
        cjpg->cbToGet -= cnt;
    }

    rc = DosAllocMem( (void*)&cjpg->pInbuf, cjpg->cbInbuf,
                      PAG_COMMIT | PAG_READ | PAG_WRITE);
    if (rc) {
        camcli_debug( "JSavRot_stdio_src - DosAllocMem  rc= 0x%04x", rc);
        cjpg->pInbuf = 0;
        csr->rtn = CAMERR_MALLOC;
        break;
    }

    memcpy( cjpg->pInbuf, crd.data, cnt);

    cinfo->src = (struct jpeg_source_mgr *) (*cinfo->mem->alloc_small)
                 ((j_common_ptr) cinfo, JPOOL_PERMANENT, SIZEOF(struct jpeg_source_mgr));

    cinfo->src->init_source = JSavRot_init_source;
    cinfo->src->fill_input_buffer = JSavRot_fill_input_buffer;
    cinfo->src->skip_input_data = JSavRot_skip_input_data;
    cinfo->src->resync_to_restart = jpeg_resync_to_restart; // use default method
    cinfo->src->term_source = JSavRot_term_source;
    cinfo->src->bytes_in_buffer = cnt;
    cinfo->src->next_input_byte = cjpg->pInbuf;

} while (0);

    if (csr->rtn != PTP_RC_OK)
        JSavRot_abort((j_common_ptr)cinfo);

    return;
}

/****************************************************************************/

// Initialize source - called by jpeg_read_header
// before any data is actually read.

void        JSavRot_init_source( j_decompress_ptr cinfo)
{
    return;
}

/****************************************************************************/

// Fill the input buffer - called whenever buffer is emptied.

boolean     JSavRot_fill_input_buffer( j_decompress_ptr cinfo)

{
    uint32_t        cnt;

    cnt = JSavRot_read_input( cinfo);
    if (!cnt)
        JSavRot_abort((j_common_ptr)cinfo);

    cinfo->src->next_input_byte = ((CAMJpg *)cinfo->client_data)->pInbuf;
    cinfo->src->bytes_in_buffer = cnt;

    return (TRUE);
}

/****************************************************************************/

// Skip data --- used to skip over a potentially large amount of
// uninteresting data (such as an APPn marker).

void        JSavRot_skip_input_data( j_decompress_ptr cinfo, long num_bytes)

{
    if (num_bytes > 0) {
        while (num_bytes > (long) cinfo->src->bytes_in_buffer) {
            num_bytes -= (long) cinfo->src->bytes_in_buffer;
            (void) JSavRot_fill_input_buffer( cinfo);
        }
        cinfo->src->next_input_byte += (size_t) num_bytes;
        cinfo->src->bytes_in_buffer -= (size_t) num_bytes;
    }

    return;
}

/****************************************************************************/

// Terminate source - called by jpeg_finish_decompress
// after all data has been read.

void        JSavRot_term_source( j_decompress_ptr cinfo)

{
    CAMJpg *        cjpg = (CAMJpg *)cinfo->client_data;
    CAMSaveRot *    csr = cjpg->csr;

    while (JSavRot_read_input( cinfo))
        if (csr->rtn != PTP_RC_OK)
            JSavRot_abort((j_common_ptr)cinfo);

    return;
}

/****************************************************************************/

uint32_t    JSavRot_read_input( j_decompress_ptr cinfo)
{

    uint32_t        ulRtn = 0;
    uint32_t        cnt;
    CAMJpg *        cjpg = (CAMJpg *)cinfo->client_data;
    CAMSaveRot *    csr = cjpg->csr;

do {
    if (!cjpg->cbToGet)
        break;

    if (*csr->pStop) {
        csr->rtn = CancelRequest( csr->thisCam->hCam);
        if (!csr->rtn)
            csr->rtn = CAMERR_REQCANCELLED;
        break;
    }

    if (cjpg->cbToGet <= cjpg->cbInbuf) {
        cnt = cjpg->cbToGet;
        cnt++;
    }
    else
        cnt = cjpg->cbInbuf;

    csr->rtn = ReadObjectData( csr->thisCam->hCam, &cnt, cjpg->pInbuf);
    if (csr->rtn != PTP_RC_OK)
        break;

    cjpg->cbToGet -= cnt;
    if (cjpg->cbToGet == 0) {
        csr->rtn = ReadObjectEnd( csr->thisCam->hCam);
        if (csr->rtn != PTP_RC_OK)
            break;
    }

    ulRtn = cnt;

} while (0);

    return (ulRtn);
}

/****************************************************************************/
/****************************************************************************/

// Prepare for output to a file.

void        JSavRot_stdio_dest( j_compress_ptr cinfo)

{
    uint32_t        rc;
    CAMJpg *        cjpg = (CAMJpg *)cinfo->client_data;
    CAMSaveRot *    csr = cjpg->csr;

do {
    rc = DosAllocMem( (void*)&cjpg->pOutbuf, OUTPUT_BUF_SIZE,
                       	    PAG_COMMIT | PAG_READ | PAG_WRITE);
    if (rc) {
        camcli_debug( "JSavRot_stdio_dest - DosAllocMem  rc= 0x%04x", csr->rtn);
        cjpg->pOutbuf = 0;
        csr->rtn = CAMERR_MALLOC;
        break;
    }

    cinfo->dest = (struct jpeg_destination_mgr *) (*cinfo->mem->alloc_small)
                  ((j_common_ptr) cinfo, JPOOL_PERMANENT, SIZEOF(struct jpeg_destination_mgr));

    cinfo->dest->init_destination = JSavRot_init_destination;
    cinfo->dest->empty_output_buffer = JSavRot_empty_output_buffer;
    cinfo->dest->term_destination = JSavRot_term_destination;
    cinfo->dest->next_output_byte = cjpg->pOutbuf;
    cinfo->dest->free_in_buffer = OUTPUT_BUF_SIZE;

} while (0);

    if (csr->rtn != PTP_RC_OK) 
        JSavRot_abort((j_common_ptr)cinfo);

    return;
}

/****************************************************************************/

// Initialize destination --- called by jpeg_start_compress
// before any data is actually written.

void        JSavRot_init_destination( j_compress_ptr cinfo)
{
    return;
}

/****************************************************************************/

// Empty the output buffer - called whenever buffer fills up.

boolean     JSavRot_empty_output_buffer( j_compress_ptr cinfo)
{
    uint32_t        rc;
    uint32_t        wrote;
    CAMJpg *        cjpg = (CAMJpg *)cinfo->client_data;
    CAMSaveRot *    csr = cjpg->csr;

    rc = DosWrite( csr->hFile, cjpg->pOutbuf, OUTPUT_BUF_SIZE, (PULONG)&wrote);
    if (rc) {
        camcli_debug( "JSavRot_empty_output_buffer - DosWrite  rc= 0x%04x", rc);
        csr->rtn = CAMERR_WRITEFAILED;
        JSavRot_abort((j_common_ptr)cinfo);
    }

    cinfo->dest->next_output_byte = cjpg->pOutbuf;
    cinfo->dest->free_in_buffer = OUTPUT_BUF_SIZE;

    return TRUE;
}

/****************************************************************************/

// Terminate destination --- called by jpeg_finish_compress
// after all data has been written.

void        JSavRot_term_destination( j_compress_ptr cinfo)
{
    uint32_t        rc;
    uint32_t    	wrote;
    uint32_t    	cnt;
    CAMJpg *        cjpg = (CAMJpg *)cinfo->client_data;
    CAMSaveRot *    csr = cjpg->csr;

    cnt = OUTPUT_BUF_SIZE - cinfo->dest->free_in_buffer;

    // Write any data remaining in the buffer
    if (cnt > 0) {
        rc = DosWrite( csr->hFile, cjpg->pOutbuf, cnt, (PULONG)&wrote);
        if (rc) {
            camcli_debug( "term_destination - DosWrite  rc= 0x%04x", rc);
            csr->rtn = CAMERR_WRITEFAILED;
            JSavRot_abort((j_common_ptr)cinfo);
        }
    }

    return;
}

/****************************************************************************/
/****************************************************************************/

// Error exit handler: must not return to caller.

void        JSavRot_error_exit (j_common_ptr cinfo)

{
    CAMJpg *    cjpg = (CAMJpg *)cinfo->client_data;

    printf( "JSavRot_error_exit called\n");

    // Always display the message
    (*cinfo->err->output_message) (cinfo);

    cjpg->csr->rtn = CAMERR_LIBJPEG;
    JSavRot_abort( cinfo);

    exit( EXIT_FAILURE);
}

/****************************************************************************/

void        JSavRot_abort( j_common_ptr cinfo)

{
    CAMJpg *    cjpg = (CAMJpg *)cinfo->client_data;

    printf( "JSavRot_abort called\n");

    if (cjpg->cbToGet)
        CancelRequest( cjpg->csr->thisCam->hCam);

    longjmp( cjpg->jmpbuf, 1);

    exit( EXIT_FAILURE);
}

/****************************************************************************/
/****************************************************************************/

uint32_t    JpgDecodeThumbnail( CAMJThumb * cjt)

{
    uint32_t                        cbRow;
    uint32_t                        cntRows;
    uint32_t                        ctr;
    JSAMPARRAY                      buffer;
    char *                          ptr;
    struct jpeg_decompress_struct   srcinfo;
    struct jpeg_error_mgr           jsrcerr;
    CAMJThumbInt                    cjpg;

do {
    memset( &cjpg, 0, sizeof(CAMJThumbInt));
    memset( cjpg.eoi, M_EOI, sizeof(cjpg.eoi));
    cjpg.cjt = cjt;
    if (setjmp( cjpg.jmpbuf))
        break;

    // Initialize the JPEG decompression object with custom error handling.
    srcinfo.client_data = &cjpg;
    srcinfo.err = jpeg_std_error( &jsrcerr);
    jsrcerr.error_exit = &JThumb_error_exit;
    jpeg_create_decompress( &srcinfo);

    // Specify data source for decompression
    JThumb_stdio_src( &srcinfo);

    // Read file header then start decompression
    jpeg_read_header( &srcinfo, TRUE);
    jpeg_start_decompress( &srcinfo);

    cbRow = (srcinfo.output_width * srcinfo.output_components + 3) & ~3;
    cntRows = srcinfo.rec_outbuf_height;
    buffer = (*srcinfo.mem->alloc_sarray)
		      ( (j_common_ptr)&srcinfo, JPOOL_IMAGE, cbRow, cntRows);

    cjt->cxDst = srcinfo.output_width;
    cjt->cyDst = srcinfo.output_height;
    cjt->cbPix = srcinfo.output_components;
    cjt->cbDst = cbRow * srcinfo.output_height;
    cjt->pDst = malloc( cjt->cbDst);
    if (!cjt->pDst) {
        cjpg.rtn = CAMERR_MALLOC;
        break;
    }

    ptr = &cjt->pDst[cjt->cbDst - cbRow];

    while (srcinfo.output_scanline < srcinfo.output_height) {
        jpeg_read_scanlines( &srcinfo, buffer, cntRows);
        for (ctr = 0; ctr < cntRows; ctr++, ptr -= cbRow)
            memcpy( ptr, buffer[ctr], cbRow);
    }

    // Finish decompression and release memory
    jpeg_finish_decompress( &srcinfo);

} while (0);

    jpeg_destroy_decompress( &srcinfo);

    if (cjpg.rtn && cjt->pDst) {
        free( cjt->pDst);
        cjt->pDst = 0;
    }

    // All done
    return (cjpg.rtn);
}

/****************************************************************************/

// Setup data source manager.  For thumbnails, we already have the
// entire image, so most of the methods are no-ops.  We give the library
// what we have & if it needs more, we'll abort.

void        JThumb_stdio_src( j_decompress_ptr cinfo)

{
    CAMJThumbInt *  cjpg = (CAMJThumbInt *)cinfo->client_data;

    cinfo->src = (struct jpeg_source_mgr *) (*cinfo->mem->alloc_small)
                 ((j_common_ptr) cinfo, JPOOL_PERMANENT, SIZEOF(struct jpeg_source_mgr));

    cinfo->src->init_source = JThumb_init_source;
    cinfo->src->fill_input_buffer = JThumb_fill_input_buffer;
    cinfo->src->skip_input_data = JThumb_skip_input_data;
    cinfo->src->resync_to_restart = jpeg_resync_to_restart; // use default method
    cinfo->src->term_source = JThumb_term_source;
    cinfo->src->bytes_in_buffer = cjpg->cjt->cbSrc;
    cinfo->src->next_input_byte = cjpg->cjt->pSrc;

    return;
}

/****************************************************************************/

// Initialize source - called by jpeg_read_header before
// any data is actually read.  For thumbnails, it's a no-op.

void        JThumb_init_source( j_decompress_ptr cinfo)
{
    return;
}

/****************************************************************************/

// Fill the input buffer - called whenever buffer is emptied.
// This should never be called because we started out with a
// complete image.  If it's called once, return some EOIs in
// the hope that the library will stop processing.  If it's
// called again, abort.

boolean     JThumb_fill_input_buffer( j_decompress_ptr cinfo)

{
    CAMJThumbInt *  cjpg = (CAMJThumbInt *)cinfo->client_data;

    if (cjpg->eoiCtr) {
        printf( "2nd call to JThumb_fill_input_buffer - aborting\n");
        cjpg->rtn = CAMERR_NEEDDATA;
        JThumb_abort( (j_common_ptr)cinfo);
    }

    printf( "1st call to JThumb_fill_input_buffer - returning EOI\n");
    cinfo->src->bytes_in_buffer = sizeof(cjpg->eoi);
    cinfo->src->next_input_byte = cjpg->eoi;
    cjpg->eoiCtr++;

    return (TRUE);
}

/****************************************************************************/

// Skip data --- used to skip over a potentially large amount of
// uninteresting data (such as an APPn marker).

void        JThumb_skip_input_data( j_decompress_ptr cinfo, long num_bytes)

{
    if (num_bytes > 0) {
        while (num_bytes > (long) cinfo->src->bytes_in_buffer) {
            num_bytes -= (long) cinfo->src->bytes_in_buffer;
            JThumb_fill_input_buffer( cinfo);
        }
        cinfo->src->next_input_byte += (size_t)num_bytes;
        cinfo->src->bytes_in_buffer -= (size_t)num_bytes;
    }

    return;
}

/****************************************************************************/

// Terminate source - called by jpeg_finish_decompress after
// all data has been read.  For thumbnails, it's a no-op.

void        JThumb_term_source( j_decompress_ptr cinfo)

{
    return;
}

/****************************************************************************/

// Error exit handler: must not return to caller.

void        JThumb_error_exit( j_common_ptr cinfo)

{
    CAMJThumbInt *  cjpg = (CAMJThumbInt *)cinfo->client_data;

    printf( "JThumb_error_exit called\n");

    // Always display the message
    (*cinfo->err->output_message)( cinfo);

    cjpg->rtn = CAMERR_LIBJPEG;
    JThumb_abort( cinfo);

    exit( EXIT_FAILURE);
}

/****************************************************************************/

void        JThumb_abort( j_common_ptr cinfo)

{
    CAMJThumbInt *  cjpg = (CAMJThumbInt *)cinfo->client_data;

    printf( "JThumb_abort called\n");
    longjmp( cjpg->jmpbuf, 1);

    exit( EXIT_FAILURE);
}

/****************************************************************************/
/****************************************************************************/

uint32_t    JpgFindOrientation( char * buf, uint32_t count, uint32_t* orientation)

{
    uint32_t    ulRtn;
    uint32_t    fLE;
    uint32_t    cntTags;
    char *      pHdr;
    char *      ptr;
    char *      end = &buf[count];

do {
    *orientation = 0;

    ulRtn = JpgFindExifTiffHdr( buf, end, &pHdr);
    if (ulRtn)
        break;

    ulRtn = JpgFind0thIfd( pHdr, end, &ptr, &fLE);
    if (ulRtn)
        break;

    if (&ptr[1] >= end) {
        ulRtn = CAMERR_NEEDDATA;
        break;
    }

    // get the number of tags
    if (fLE)
        cntTags = *((uint16_t*)ptr);
    else
        cntTags = (ptr[0] << 8) + ptr[1];
    ptr += 2;

    // go through the array of 12-byte tags looking for the orientation tag
    // (this is supposed to be in ascending order but I don't trust "them")

    ulRtn = CAMERR_NOORIENT;
    while (cntTags && ptr+12 <= end) {
        if ((fLE && *((uint16_t*)ptr) == 0x0112) || *((uint16_t*)ptr) == 0x1201) {
            if (fLE)
                *orientation = ptr[8] + (ptr[9] << 8);
            else
                *orientation = (ptr[8] << 8) + ptr[9];

            // if the orientation is between 1 & 8, signal success,
            // otherwise, clear the value
            if (*orientation >= 1 && *orientation <= 8)
                ulRtn = 0;
            else
                *orientation = 0;

            break;
        }

        cntTags--;
        ptr += 12;
    }

} while (0);

    return (ulRtn);
}

/****************************************************************************/

uint32_t    JpgFindExifTiffHdr( char * buf, char * end, char** ppHdr)

{
    uint32_t    ulRtn = CAMERR_NEEDDATA;
    uint32_t    lth;
    char        marker;
    char *      ptr = buf;

do {
    // the JPEG tag must be the first two bytes
    if (*ptr++ != (char)0xff ||
        *ptr++ != (char)M_SOI) {
        ulRtn = CAMERR_BADFORMAT;
        break;
    }

    // the Exif tag is _supposed_ to appear immediately after SOI,
    // but then, who actually follows the standards?
    while (ptr < end) {
        // skip over junk
        while (ptr < end && *ptr != (char)0xff)
            ptr++;
        if (ptr >= end)
            break;

        // skip over multiple ff's
        while (ptr < end && *ptr == (char)0xff)
            ptr++;
        if (ptr >= end)
            break;

        // if we've gone too far or don't have any more data, exit
        marker = *ptr++;
        if (&ptr[1] >= end)
            break;
        if (marker == (char)M_EOI || marker == (char)M_SOS) {
            ulRtn = CAMERR_NOEXIF;
            break;
        }

        // get the marker length (which has to be at least 2)
        lth = (ptr[0] << 8) + ptr[1];
        if (lth < 2) {
            ulRtn = CAMERR_BADFORMAT;
            break;
        }

        ptr += 2;
        lth -= 2;

        // if this is an APP1 that's minimally long enough
        // and says "Exif", we've found our marker
        if (&ptr[5] >= end)
            break;

        if (marker == (char)M_APP1 && lth >= 6 &&
            !memcmp( ptr, "Exif\0", 6)) {
            ptr += 6;
            ulRtn = 0;
            break;
        }

        // advance to next marker
        ptr += lth;
    }

} while (0);

    *ppHdr = ptr;

    return (ulRtn);
}

/****************************************************************************/

uint32_t    JpgFind0thIfd( char * hdr, char * end, char** ppIFD, uint32_t* fLE)

{
    uint32_t    ulRtn = CAMERR_BADFORMAT;
    uint32_t    offs;
    char *      ptr = hdr;

do {

    // is there enough data to examine the 8-byte TIFF header?
    if (&ptr[7] >= end) {
        ulRtn = CAMERR_NEEDDATA;
        break;
    }

    // get the byte order
    if (ptr[0] == 'I' && ptr[1] == 'I')
        *fLE = TRUE;
    else
    if (ptr[0] == 'M' && ptr[1] == 'M')
        *fLE = FALSE;
    else
        break;

    // this is always '0x2a'
    ptr += 2;
    if (*fLE) {
        if (*((uint16_t*)ptr) != 0x002a)
            break;
    }
    else
        if (*((uint16_t*)ptr) != 0x2a00)
            break;

    // advance to the beginning of the IFD
    ptr += 2;
    if (*fLE)
        offs = *((uint32_t*)ptr);
    else
        offs = ((ptr[0] << 24) + (ptr[1] << 16) + (ptr[2] << 8) + ptr[3]);

    if (offs < 8 || offs >= 0x10000)
        break;

    ptr += offs - 4;

    if (ptr >= end)
        ulRtn = CAMERR_NEEDDATA;
    else
        ulRtn = 0;

} while (0);

    *ppIFD = ptr;

    return (ulRtn);
}

/****************************************************************************/
/****************************************************************************/

