/****************************************************************************/
// camthumb.c
//
//  - fetch thumbnails & Exif info from camera, create thumbnail bitmaps
//
/****************************************************************************/

/* ***** BEGIN LICENSE BLOCK *****
 * Version: MPL 1.1
 *
 * The contents of this file are subject to the Mozilla Public License Version
 * 1.1 (the "License"); you may not use this file except in compliance with
 * the License. You may obtain a copy of the License at
 * http://www.mozilla.org/MPL/
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 * for the specific language governing rights and limitations under the
 * License.
 *
 * The Original Code is Cameraderie.
 *
 * The Initial Developer of the Original Code is Richard L. Walsh.
 * 
 * Portions created by the Initial Developer are Copyright (C) 2006,2007
 * the Initial Developer. All Rights Reserved.
 *
 * Contributor(s):
 *
 * ***** END LICENSE BLOCK ***** */

/****************************************************************************/

#include "camapp.h"
#include "camusb.h"

/****************************************************************************/

typedef struct _CAMThumb {
    HWND        hReply;
    CAMPtr      thisCam;
    BOOL        fRotate;
    CAMRecPtr   pcr[1];
} CAMThumb;

typedef CAMThumb *CAMThumbPtr;

/****************************************************************************/

void            FetchAllThumbnails( HWND hwnd);
void            FetchOneThumbnail( HWND hwnd, MPARAM mp1, MPARAM mp2);
void            ThumbnailThreadProc( void * camthumbptr);
void            SetOrientation( CAMPtr thisCam, CAMRecPtr pcr);
HBITMAP         CreateThumbnailBmp( CAMRecPtr pcr, char * pBuf);

/****************************************************************************/

void        FetchAllThumbnails( HWND hwnd)

{
    BOOL            fRtn = FALSE;
    ULONG           ctr;
    HWND            hCnr;
    CAMRecPtr       pcr;
    CAMRecPtr *     ppcr;
    CAMThumbPtr     pct = 0;
    char *          pErr = 0;
    DATETIME        dt;
    CNRINFO         ci;

do {
    DosGetDateTime( &dt);

    hCnr = WinWindowFromID( hwnd, FID_CLIENT);

    if (!WinSendMsg( hCnr, CM_QUERYCNRINFO, (MP)&ci, (MP)sizeof(CNRINFO))) {
        pErr = "CM_QUERYCNRINFO";
        break;
    }

    ctr = ci.cRecords * sizeof(CAMRecPtr) + sizeof(CAMThumb);
    pct = (CAMThumbPtr)malloc( ctr);
    if (!pct) {
        pErr = "malloc";
        break;
    }
    memset( pct, 0, ctr);

    // fill in useful info
    pct->hReply  = hwnd;
    pct->thisCam = (CAMPtr)GetCamPtr();
    pct->fRotate = (QueryRotateOpts() & CAM_ROTATEONLOAD);

    ctr = 0;
    ppcr = pct->pcr;
    pcr = (CAMRecPtr)WinSendMsg( hCnr, CM_QUERYRECORD, (MP)0,
                                 MPFROM2SHORT( CMA_FIRST, CMA_ITEMORDER));

    while (pcr && pcr != (CAMRecPtr)-1 && ctr < ci.cRecords) {
        if (!pcr->bmp && pcr->tnsize) {
            *ppcr++ = pcr;
            ctr++;
        }
        pcr = (CAMRecPtr)WinSendMsg( hCnr, CM_QUERYRECORD, (MP)pcr,
                                     MPFROM2SHORT( CMA_NEXT, CMA_ITEMORDER));
    }

    if (!ctr)
        break;

    if (pcr) {
        if (pcr == (CAMRecPtr)-1)
            pErr = "CM_QUERYRECORD";
        else
            pErr = "record overflow";
        break;
    }

    // start the thread
    if (_beginthread( ThumbnailThreadProc, 0, 0xF000, pct) == -1) {
        pErr = "_beginthread";
        break;
    }

    fRtn = TRUE;

} while (0);

    if (!fRtn) {
        if (pErr)
            printf( "FetchAllThumbnails - %s failed\n", pErr);
        if (pct)
            free( pct);
    }
    else
    printf( "fetch begin= %02hhd:%02hhd:%02hhd.%02hhd\n",
            dt.hours, dt.minutes, dt.seconds, dt.hundredths);


    return;
}

/****************************************************************************/

void        FetchOneThumbnail( HWND hwnd, MPARAM mp1, MPARAM mp2)

{
    CAMRecPtr   pcr = (CAMRecPtr)mp1;
    DATETIME    dt;

    if (pcr == (CAMRecPtr)-1) {
        DosGetDateTime( &dt);
        printf( "fetch   end= %02hhd:%02hhd:%02hhd.%02hhd - rc= 0x%x\n",
                dt.hours, dt.minutes, dt.seconds, dt.hundredths, (int)mp2);
        return;
    }

    if (pcr && pcr->bmp) {
        WinSendDlgItemMsg( hwnd, FID_CLIENT, CM_INVALIDATERECORD, (MP)&pcr,
                           MPFROM2SHORT( 1, (CMA_NOREPOSITION | CMA_ERASE)));
        if (pcr->mrc.flRecordAttr & CRA_CURSORED)
            WinPostMsg( hwnd, CAMMSG_SHOWTHUMB, (MP)pcr, 0);
    }

    return;
}

/****************************************************************************/
// everything below runs on a secondary thread
/****************************************************************************/

void        ThumbnailThreadProc( void * camthumbptr)

{
    uint32_t        rc = 0;
    uint32_t        errCtr = 0;
    char *          pBuf = 0;
    CAMThumbPtr     pct = (CAMThumbPtr)camthumbptr;
    CAMRecPtr *     ppcr = pct->pcr;
    CAMRecPtr       pcr;

do{

    printf( "Entering ThumbnailThreadProc\n");

    rc = DosRequestMutexSem( pct->thisCam->hmtxCam, CAM_MUTEXWAIT);
    if (rc)
        break;

    while (*ppcr) {
        pcr = *ppcr;

        // get the thumb from the camera
        rc = GetThumb( pct->thisCam->hCam, pcr->hndl, &pBuf);
        if (rc) {
            printf( "ThumbnailThreadProc - GetThumb #1 - handle= %d  rc= 0x%x\n",
                    (int)pcr->hndl, (int)rc);
            if (rc == CAMERR_NULLCAMERAPTR)
                break;

            rc = ClearStall( pct->thisCam->hCam);
            if (rc == CAMERR_NULLCAMERAPTR)
                break;

            rc = GetThumb( pct->thisCam->hCam, pcr->hndl, &pBuf);
            if (rc) {
                printf( "ThumbnailThreadProc - GetThumb #2 - handle= %d  rc= 0x%x\n",
                        (int)pcr->hndl, (int)rc);
                if (rc == CAMERR_NULLCAMERAPTR)
                    break;

                ClearStall( pct->thisCam->hCam);
                if (++errCtr > 2)
                    break;
            }
        }

        if (!rc) {
            errCtr = 0;

            // if rotate-on-load is set, get the
            // Exif orientation info from the main image
            if (pct->fRotate &&
                pcr->fmtnbr == PTP_OFC_EXIF_JPEG &&
                pcr->size >= 500)
                    SetOrientation( pct->thisCam, pcr);

            // create the bitmap
            pcr->bmp = CreateThumbnailBmp( pcr, pBuf);
        }

        // free the memory allocated by GetThumb()
        if (pBuf) {
            free( pBuf);
            pBuf = 0;
        }

        // let the main thread know we finished processing this record
        if (pcr->bmp &&
            !WinPostMsg( pct->hReply, CAMMSG_FETCHTHUMBS, (MP)pcr, 0))
            printf( "ThumbnailThreadProc - WinPostMsg failed\n");

        ppcr++;
    }

    DosReleaseMutexSem( pct->thisCam->hmtxCam);


} while (0);

    WinPostMsg( pct->hReply, CAMMSG_FETCHTHUMBS, (MP)-1, (MP)rc);

    free( pct);
    printf( "Exiting ThumbnailThreadProc - rc= 0x%x  errCtr= %d\n",
            (int)rc, (int)errCtr);

    return;
}

/***************************************************************************/

// gets the Exif orientation info from the main image, then sets the
// orientation flags on the container record;  it does so by getting
// the first 500 bytes of the image which almost certainly should
// contain the Exif header - if not, too bad

void        SetOrientation( CAMPtr thisCam, CAMRecPtr pcr)

{
    uint32_t        rc = 0;
    uint32_t        cnt;
    uint32_t        orientation;
    CAMReadData     crd;
    char *          pbuf = crd.data;

do {
    cnt = sizeof(crd.data);

    // there are 3 ways to get the first 500 bytes;
    // the one we use depends on what the camera supports
    if (thisCam->camFlags & CAM_PARTIAL_STANDARD)
        rc = GetPartialObject( thisCam->hCam, pcr->hndl, 0,
                               &cnt, &pbuf);
    else
    if (thisCam->camFlags & CAM_PARTIAL_CANON)
        rc = CanonGetPartialObject( thisCam->hCam, pcr->hndl, 0,
                                    &cnt, 0, &pbuf);
    else {
        // this one is ugly, start fetching the entire image,
        // then kill it after the first read (it works, though)
        rc = ReadObjectBegin( thisCam->hCam, pcr->hndl, &cnt, &crd);
        if (rc == PTP_RC_OK)
            rc = CancelRequest( thisCam->hCam);
    }

    // something went wrong
    if (rc || cnt < 500) {
        camcli_error( "GetPartial failed for handle %ld - count= %ld  rc= 0x%lx",
                      pcr->hndl, cnt, rc);
        if (rc)
            ClearStall( thisCam->hCam);
        break;
    }

    // parse the Exif info looking for orientation info
    rc = JpgFindOrientation( crd.data, cnt, &orientation);
    if (rc) {
        camcli_error( "FindOrientation failed for handle %ld - rc= 0x%lx",
                      pcr->hndl, rc);
        break;
    }

    // set the record's orientation flag
    switch (orientation) {
        case 1:
        case 2:
            pcr->status = STAT_SETROT0(pcr->status);
            break;

        case 3:
        case 4:
            pcr->status = STAT_SETROT180(pcr->status);
            break;

        case 5:
        case 6:
            pcr->status = STAT_SETROT270(pcr->status);
            break;

        case 7:
        case 8:
            pcr->status = STAT_SETROT90(pcr->status);
            break;

        default:
            camcli_error( "Invalid orientation flag for handle %ld - orientation= %ld",
                          pcr->hndl, orientation);
            pcr->status = STAT_SETROT0(pcr->status);
            break;
    }

} while (0);

    return;
}

/****************************************************************************/

HBITMAP     CreateThumbnailBmp( CAMRecPtr pcr, char * pBuf)

{
    HBITMAP             hBmp = 0;
    uint32_t            ctr;
    HDC                 hdc = 0;
    HPS                 hps = 0;
    char *              pErr = 0;
    BITMAPINFOHEADER2 * pbmih = 0;
    RGB2 *              pRGB;
    SIZEL               sizl;
    CAMJThumb           cjt;

do {
    // decode the JPEG data from the camera
    memset( &cjt, 0, sizeof(cjt));
    cjt.pSrc = pBuf;
    cjt.cbSrc = pcr->tnsize;
    ctr = JpgDecodeThumbnail( &cjt);
    if (ctr) {
        printf( "JpgDecodeThumbnail - rc= 0x%x\n", (int)ctr);
        break;
    }

    // allocate a BITMAPINFOHEADER2 / BITMAPINFO2
    // suitable for 24-bit color or 8-bit grayscale
    ctr = sizeof(BITMAPINFOHEADER2);
    if (cjt.cbPix == 1)
        ctr += 256 * sizeof(RGB2);
        
    pbmih = (BITMAPINFOHEADER2*)malloc( ctr);
    memset( pbmih, 0, ctr);

    // if grayscale, create a default colormap
    if (cjt.cbPix == 1)
        for (ctr = 0, pRGB = (RGB2*)&pbmih[1]; ctr < 256; ctr++, pRGB++) {
            pRGB->bBlue  = ctr;
            pRGB->bGreen = ctr;
            pRGB->bRed   = ctr;
        }

    // init the BMIH
    pbmih->cbFix         = sizeof(BITMAPINFOHEADER2);
    pbmih->cx            = cjt.cxDst;
    pbmih->cy            = cjt.cyDst;
    pbmih->cPlanes       = 1;
    pbmih->cBitCount     = cjt.cbPix * 8;
    pbmih->ulCompression = BCA_UNCOMP;
    pbmih->cbImage       = cjt.cbDst;

    // create a memory DC
    hdc = DevOpenDC( 0, OD_MEMORY, "*", 0L, 0, 0);
    if (!hdc) {
        pErr = "DevOpenDC";
        break;
    }

    // create a square PS using the longer side so we can
    // reuse the PS if we have to rotate the bitmap
    sizl.cx = ((pbmih->cx >= pbmih->cy) ? pbmih->cx : pbmih->cy);
    sizl.cy = sizl.cx;

    hps = GpiCreatePS( 0, hdc, &sizl, PU_PELS | GPIF_DEFAULT | GPIT_MICRO | GPIA_ASSOC);
    if (!hps) {
        pErr = "GpiCreatePS";
        break;
    }

    // create a bitmap from the data
    hBmp = GpiCreateBitmap( hps, pbmih, CBM_INIT, cjt.pDst, (BITMAPINFO2*)pbmih);
    if (!hBmp) {
        pErr = "GpiCreateBitmap";
        break;
    }

    // if the bitmap should be rotated, replace the original one
    if (STAT_GETROT(pcr->status)) {
        HBITMAP hBmpRot = RotateBitmap( hps, hBmp, pbmih,
                                        STAT_GETROT(pcr->status));
        if (hBmpRot) {
            GpiDeleteBitmap( hBmp);
            hBmp = hBmpRot;
        }
    }

} while (0);

    if (pErr)
        printf( "CreateThumbnailBmp - %s failed\n", pErr);

    if (hps) {
        GpiSetBitmap( hps, 0);
        GpiDestroyPS( hps);
    }
    if (hdc)
        DevCloseDC( hdc);
    if (pbmih)
        free( pbmih);
    if (cjt.pDst)
        free( cjt.pDst);

    return (hBmp);
}

/****************************************************************************/

